@@ -1,4 +1,4 @@
-/* Copyright (c) 2011-2013 Yubico AB
+/* Copyright (c) 2011-2014 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,6 +34,7 @@
 #else
 
 #include <pwd.h>
+#include <stdio.h>
 
 #ifdef HAVE_SECURITY_PAM_APPL_H
 #include <security/pam_appl.h>
@@ -49,11 +50,12 @@ struct _ykpam_privs {
   gid_t saved_egid;
   gid_t *saved_groups;
   int saved_groups_length;
+  FILE *debug_file;
 };
 
 #define PAM_MODUTIL_DEF_PRIVS(n) \
   gid_t n##_saved_groups[SAVED_GROUPS_MAX_LEN]; \
-  struct _ykpam_privs n = {-1, -1, n##_saved_groups, SAVED_GROUPS_MAX_LEN}
+  struct _ykpam_privs n = {-1, -1, n##_saved_groups, SAVED_GROUPS_MAX_LEN, cfg->debug_file}
 
 int pam_modutil_drop_priv(pam_handle_t *, struct _ykpam_privs *, struct passwd *);
 int pam_modutil_regain_priv(pam_handle_t *, struct _ykpam_privs *);
