@@ -38,7 +38,7 @@
 #include <sys/stat.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <pwd.h>
+#include <glob.h>
 #include <unistd.h>
 
 #include "util.h"
@@ -53,16 +53,15 @@
 #endif /* HAVE_CR */
 
 int
-get_user_cfgfile_path(const char *common_path, const char *filename, const char *username, char **fn)
+get_user_cfgfile_path(const char *common_path, const char *filename, const struct passwd *user, char **fn)
 {
   /* Getting file from user home directory, e.g. ~/.yubico/challenge, or
    * from a system wide directory.
    *
    * Format is hex(challenge):hex(response):slot num
    */
-  struct passwd *p;
   char *userfile;
-  int len;
+  size_t len;
 
   if (common_path != NULL) {
     len = strlen(common_path) + 1 + strlen(filename) + 1;
@@ -76,23 +75,20 @@ get_user_cfgfile_path(const char *common_path, const char *filename, const char
 
   /* No common path provided. Construct path to user's ~/.yubico/filename */
 
-  p = getpwnam (username);
-  if (!p)
-    return 0;
-
-  len = strlen(p->pw_dir) + 9 + strlen(filename) + 1;
+  len = strlen(user->pw_dir) + 9 + strlen(filename) + 1;
   if ((userfile = malloc(len)) == NULL) {
     return 0;
   }
-  snprintf(userfile, len, "%s/.yubico/%s", p->pw_dir, filename);
+  snprintf(userfile, len, "%s/.yubico/%s", user->pw_dir, filename);
   *fn = userfile;
   return 1;
 }
 
 
 /*
- * This function will look for users name with valid user token id. It
- * will returns -2 if the user is unknown, -1 if the token do not match the user line, 0 for internal failure and 1 for success.
+ * This function will look for users name with valid user token id.
+ *
+ * Returns one of AUTH_FOUND, AUTH_NOT_FOUND, AUTH_NO_TOKENS, AUTH_ERROR.
  *
  * File format is as follows:
  * <user-name>:<token_id>:<token_id>
@@ -103,32 +99,33 @@ int
 check_user_token (const char *authfile,
 		  const char *username,
 		  const char *otp_id,
-		  int verbose)
+		  int verbose,
+                  FILE *debug_file)
 {
   char buf[1024];
   char *s_user, *s_token;
-  int retval = 0;
+  int retval = AUTH_ERROR;
   int fd;
   struct stat st;
   FILE *opwfile;
 
-  fd = open(authfile, O_RDONLY, 0);
+  fd = open(authfile, O_RDONLY | O_CLOEXEC, 0);
   if (fd < 0) {
       if(verbose)
-	  D (("Cannot open file: %s (%s)", authfile, strerror(errno)));
+	  D (debug_file, "Cannot open file: %s (%s)", authfile, strerror(errno));
       return retval;
   }
 
   if (fstat(fd, &st) < 0) {
       if(verbose)
-	  D (("Cannot stat file: %s (%s)", authfile, strerror(errno)));
+	  D (debug_file, "Cannot stat file: %s (%s)", authfile, strerror(errno));
       close(fd);
       return retval;
   }
 
   if (!S_ISREG(st.st_mode)) {
       if(verbose)
-	  D (("%s is not a regular file", authfile));
+	  D (debug_file, "%s is not a regular file", authfile);
       close(fd);
       return retval;
   }
@@ -136,36 +133,43 @@ check_user_token (const char *authfile,
   opwfile = fdopen(fd, "r");
   if (opwfile == NULL) {
       if(verbose)
-	  D (("fdopen: %s", strerror(errno)));
+	  D (debug_file, "fdopen: %s", strerror(errno));
       close(fd);
       return retval;
   }
 
-  retval = -2;
+  retval = AUTH_NO_TOKENS;
   while (fgets (buf, 1024, opwfile))
     {
       char *saveptr = NULL;
       if (buf[strlen (buf) - 1] == '\n')
 	buf[strlen (buf) - 1] = '\0';
+      if (buf[0] == '#') {
+          /* This is a comment and we may skip it. */
+          if(verbose)
+              D (debug_file, "Skipping comment line: %s", buf);
+          continue;
+      }
       if(verbose)
-	  D (("Authorization line: %s", buf));
+	  D (debug_file, "Authorization line: %s", buf);
       s_user = strtok_r (buf, ":", &saveptr);
       if (s_user && strcmp (username, s_user) == 0)
 	{
 	  if(verbose)
-	      D (("Matched user: %s", s_user));
-      retval = -1; //We found at least one line for the user
+	      D (debug_file, "Matched user: %s", s_user);
+      retval = AUTH_NOT_FOUND; /* We found at least one line for the user */
 	  do
 	    {
 	      s_token = strtok_r (NULL, ":", &saveptr);
 	      if(verbose)
-		  D (("Authorization token: %s", s_token));
-	      if (s_token && strcmp (otp_id, s_token) == 0)
+		  D (debug_file, "Authorization token: %s", s_token);
+	      if (s_token && otp_id && strcmp (otp_id, s_token) == 0)
 		{
 		  if(verbose)
-		      D (("Match user/token as %s/%s", username, otp_id));
-		  fclose (opwfile);
-		  return 1;
+		      D (debug_file, "Match user/token as %s/%s", username, otp_id);
+
+		  fclose(opwfile);
+		  return AUTH_FOUND;
 		}
 	    }
 	  while (s_token != NULL);
@@ -183,8 +187,14 @@ int generate_random(void *buf, int len)
 {
 	FILE *u;
 	int res;
+	int fd;
 
-	u = fopen("/dev/urandom", "r");
+	fd = open("/dev/urandom", O_RDONLY | O_CLOEXEC);
+	if (fd < 0) {
+		return -1;
+	}
+
+	u = fdopen(fd, "r");
 	if (!u) {
 		return -1;
 	}
@@ -196,7 +206,7 @@ int generate_random(void *buf, int len)
 }
 
 int
-check_firmware_version(YK_KEY *yk, bool verbose, bool quiet)
+check_firmware_version(YK_KEY *yk, bool verbose, bool quiet, FILE *debug_file)
 {
 	YK_STATUS *st = ykds_alloc();
 
@@ -206,11 +216,10 @@ check_firmware_version(YK_KEY *yk, bool verbose, bool quiet)
 	}
 
 	if (verbose) {
-		printf("Firmware version %d.%d.%d\n",
+		D(debug_file, "YubiKey Firmware version: %d.%d.%d\n",
 		       ykds_version_major(st),
 		       ykds_version_minor(st),
 		       ykds_version_build(st));
-		fflush(stdout);
 	}
 
 	if (ykds_version_major(st) < 2 ||
@@ -282,7 +291,72 @@ int challenge_response(YK_KEY *yk, int slot,
 }
 
 int
-get_user_challenge_file(YK_KEY *yk, const char *chalresp_path, const char *username, char **fn)
+check_user_challenge_file(const char *chalresp_path, const struct passwd *user, FILE *debug_file)
+{
+  /*
+   * This function will look for users challenge files.
+   *
+   * Returns one of AUTH_FOUND, AUTH_NOT_FOUND, AUTH_ERROR
+   */
+  size_t len;
+  int r;
+  int ret = AUTH_NOT_FOUND;
+  char *userfile = NULL;
+  char *userfile_pattern = NULL;
+  glob_t userfile_glob;
+  const char *filename = NULL;
+
+  if (! chalresp_path) {
+    filename = "challenge";
+  } else {
+    filename = user->pw_name;
+  }
+
+  /* check for userfile challenge files */
+  r = get_user_cfgfile_path(chalresp_path, filename, user, &userfile);
+  if (!r) {
+    D (debug_file, "Failed to get user cfgfile path");
+    ret = AUTH_ERROR;
+    goto out;
+  }
+
+  if (!access(userfile, F_OK)) {
+    ret = AUTH_FOUND;
+    goto out;
+  }
+
+  /* check for userfile-* challenge files */
+  len = strlen(userfile) + 2 + 1;
+  if ((userfile_pattern = malloc(len)) == NULL) {
+    D (debug_file, "Failed to allocate memory for userfile pattern: %s", strerror(errno));
+    ret = AUTH_ERROR;
+    goto out;
+  }
+  snprintf(userfile_pattern, len, "%s-*", userfile);
+
+  r = glob(userfile_pattern, 0, NULL, &userfile_glob);
+  globfree(&userfile_glob);
+  switch (r) {
+    case GLOB_NOMATCH:
+      /* No matches found, so continue */
+      break;
+    case 0:
+      ret = AUTH_FOUND;
+      goto out;
+    default:
+      D (debug_file, "Error while checking for %s challenge files: %s", userfile_pattern, strerror(errno));
+      ret = AUTH_ERROR;
+      goto out;
+  }
+
+out:
+  free(userfile_pattern);
+  free(userfile);
+  return ret;
+}
+
+int
+get_user_challenge_file(YK_KEY *yk, const char *chalresp_path, const struct passwd *user, char **fn, FILE *debug_file)
 {
   /* Getting file from user home directory, i.e. ~/.yubico/challenge, or
    * from a system wide directory.
@@ -293,28 +367,27 @@ get_user_challenge_file(YK_KEY *yk, const char *chalresp_path, const char *usern
    * the option chalresp_path can be used to point to a system-wide directory.
    */
 
-  char *filename; /* not including directory */
-  int filename_malloced = 0;
+  const char *filename = NULL; /* not including directory */
+  char *ptr = NULL;
   unsigned int serial = 0;
   int ret;
 
   if (! yk_get_serial(yk, 0, 0, &serial)) {
-    D (("Failed to read serial number (serial-api-visible disabled?)."));
+    D (debug_file, "Failed to read serial number (serial-api-visible disabled?).");
     if (! chalresp_path)
       filename = "challenge";
     else
-      filename = (char *) username;
+      filename = user->pw_name;
   } else {
     /* We have serial number */
-    int len;
     /* 0xffffffff == 4294967295 == 10 digits */
-    len = strlen(chalresp_path == NULL ? "challenge" : username) + 1 + 10 + 1;
-    if ((filename = malloc(len)) != NULL) {
-      int res = snprintf(filename, len, "%s-%u", chalresp_path == NULL ? "challenge" : username, serial);
-      filename_malloced = 1;
-      if (res < 0 || res > len) {
+    size_t len = strlen(chalresp_path == NULL ? "challenge" : user->pw_name) + 1 + 10 + 1;
+    if ((ptr = malloc(len)) != NULL) {
+      int res = snprintf(ptr, len, "%s-%u", chalresp_path == NULL ? "challenge" : user->pw_name, serial);
+      filename = ptr;
+      if (res < 0 || (unsigned long)res > len) {
 	/* Not enough space, strangely enough. */
-	free(filename);
+	free(ptr);
 	filename = NULL;
       }
     }
@@ -323,15 +396,15 @@ get_user_challenge_file(YK_KEY *yk, const char *chalresp_path, const char *usern
   if (filename == NULL)
     return 0;
 
-  ret = get_user_cfgfile_path (chalresp_path, filename, username, fn);
-  if(filename_malloced == 1) {
-    free(filename);
+  ret = get_user_cfgfile_path (chalresp_path, filename, user, fn);
+  if(ptr) {
+    free(ptr);
   }
   return ret;
 }
 
 int
-load_chalresp_state(FILE *f, CR_STATE *state, bool verbose)
+load_chalresp_state(FILE *f, CR_STATE *state, bool verbose, FILE *debug_file)
 {
   /*
    * Load the current challenge and expected response information from a file handle.
@@ -352,16 +425,16 @@ load_chalresp_state(FILE *f, CR_STATE *state, bool verbose)
    * 40 is twice the size of CR_RESPONSE_SIZE
    * (twice because we hex encode the challenge and response)
    */
-  r = fscanf(f, "v2:%126[0-9a-z]:%40[0-9a-z]:%64[0-9a-z]:%d:%d", challenge_hex, response_hex, salt_hex, &iterations, &slot);
+  r = fscanf(f, "v2:%126[0-9a-z]:%40[0-9a-z]:%64[0-9a-z]:%u:%d", challenge_hex, response_hex, salt_hex, &iterations, &slot);
   if(r == 5) {
     if (! yubikey_hex_p(salt_hex)) {
-      D(("Invalid salt hex input : %s", salt_hex));
+      D(debug_file, "Invalid salt hex input : %s", salt_hex);
       goto out;
     }
 
     if(verbose) {
-      D(("Challenge: %s, hashed response: %s, salt: %s, iterations: %d, slot: %d",
-            challenge_hex, response_hex, salt_hex, iterations, slot));
+      D(debug_file, "Challenge: %s, hashed response: %s, salt: %s, iterations: %u, slot: %d",
+            challenge_hex, response_hex, salt_hex, iterations, slot);
     }
 
     yubikey_hex_decode(state->salt, salt_hex, sizeof(state->salt));
@@ -370,12 +443,12 @@ load_chalresp_state(FILE *f, CR_STATE *state, bool verbose)
     rewind(f);
     r = fscanf(f, "v1:%126[0-9a-z]:%40[0-9a-z]:%d", challenge_hex, response_hex, &slot);
     if (r != 3) {
-      D(("Could not parse contents of chalresp_state file (%i)", r));
+      D(debug_file, "Could not parse contents of chalresp_state file (%i)", r);
       goto out;
     }
 
     if (verbose) {
-      D(("Challenge: %s, expected response: %s, slot: %d", challenge_hex, response_hex, slot));
+      D(debug_file, "Challenge: %s, expected response: %s, slot: %d", challenge_hex, response_hex, slot);
     }
 
     iterations = CR_DEFAULT_ITERATIONS;
@@ -385,17 +458,17 @@ load_chalresp_state(FILE *f, CR_STATE *state, bool verbose)
 
 
   if (! yubikey_hex_p(challenge_hex)) {
-    D(("Invalid challenge hex input : %s", challenge_hex));
+    D(debug_file, "Invalid challenge hex input : %s", challenge_hex);
     goto out;
   }
 
   if (! yubikey_hex_p(response_hex)) {
-    D(("Invalid expected response hex input : %s", response_hex));
+    D(debug_file, "Invalid expected response hex input : %s", response_hex);
     goto out;
   }
 
   if (slot != 1 && slot != 2) {
-    D(("Invalid slot input : %i", slot));
+    D(debug_file, "Invalid slot input : %i", slot);
     goto out;
   }
 
