@@ -34,15 +34,14 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/types.h>
-#include <pwd.h>
 #include <string.h>
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <fcntl.h>
 
 #include <ykpers.h>
 
-#undef DEBUG_PAM
 #include "util.h"
 
 #define ACTION_ADD_HMAC_CHALRESP	"add_hmac_chalresp"
@@ -145,7 +144,9 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, u
   unsigned int response_len;
   char *fn;
   struct passwd *p;
+  int fd;
   FILE *f = NULL;
+  struct stat st;
 
   state.iterations = iterations;
   state.slot = slot;
@@ -163,7 +164,6 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, u
   */
   
   if (!output_dir){
-      struct stat st;
       char fullpath[256];
       snprintf(fullpath, 256,"%s/.yubico",p->pw_dir);
       
@@ -186,11 +186,16 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, u
       }
   }
 
-  if (! get_user_challenge_file(yk, output_dir, p->pw_name, &fn)) {
+  if (! get_user_challenge_file(yk, output_dir, p, &fn, stdout)) {
     fprintf (stderr, "Failed getting chalresp state filename\n");
     goto out;
   }
 
+  if (stat(fn, &st) == 0) {
+    fprintf(stderr, "File %s already exists, refusing to overwrite.\n", fn);
+    goto out;
+  }
+
   if (generate_random(state.challenge, CR_CHALLENGE_SIZE)) {
     fprintf (stderr, "FAILED getting %i bytes of random data\n", CR_CHALLENGE_SIZE);
     goto out;
@@ -232,11 +237,18 @@ do_add_hmac_chalresp(YK_KEY *yk, uint8_t slot, bool verbose, char *output_dir, u
   memcpy (state.response, buf, response_len);
   state.response_len = response_len;
 
-  f = fopen (fn, "w");
-  if (! f) {
+  umask(077);
+
+  fd = open (fn, O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, S_IRUSR | S_IWUSR);
+  if (fd < 0) {
     fprintf (stderr, "Failed opening '%s' for writing : %s\n", fn, strerror (errno));
     goto out;
   }
+  f = fdopen (fd, "w");
+  if (! f) {
+    fprintf (stderr, "fdopen: %s\n", strerror (errno));
+    goto out;
+  }
 
   if (! write_chalresp_state (f, &state))
     goto out;
@@ -288,7 +300,7 @@ main(int argc, char **argv)
     if (! init_yubikey (&yk))
       goto err;
 
-    if (! check_firmware_version(yk, verbose, false))
+    if (! check_firmware_version(yk, verbose, false, stdout))
       goto err;    
 
     if (! do_add_hmac_chalresp (yk, slot, verbose, output_dir, iterations, &exit_code))
